/*+-------------------------------------------------------------------
  A Simple example of using pixel shader version 2.0.
  
  Some portions of this code, particularly the calls to
  D3DXAssembleShaderFromFile were pasted directly from the Microsoft
  DirectX 9 documentation.

  All other portions were written by Ben Landon, Copyright 2003.  This
  code is presented in the hopes that it will be useful to others, and
  it is presented without any guarantees of quality of suitability for
  any particular purpose.  Caveat emptor.  The function init_dx9 is
  written to emphasize clarity instead of being industrial stength.
  It can leak DX interfaces if it fails!

  If you use any portion of this in a tutorial, I'd appreciate being
  acknowledged, but this isn't a requirement.

*/
#include <windows.h>
#include <d3d9.h>
#include <d3dx9.h>   // for D3DXCreateTextureFromFile
#include <assert.h>


/*+-------------------------------------------------------------------
  Vertex structure

  Note that for DirectX to be happy with this structure the 
  position data has to come before the color data, which comes
  before the texture coordinates.  For more information on this
  see the documentation on flexible vertex formats. 

*/
struct Vertex
{
    FLOAT x, y, z, rhw;
    DWORD diffuse_color;
    FLOAT u, v;
};

// Use already transformed vertices so I don't have to worry about
// setting up the camera, view and projection matrices.  XYZRHW means
// that the vertices have already been transformed into viewport 
// space by the application.

#define VERTEX_FORMAT (D3DFVF_XYZRHW|D3DFVF_DIFFUSE|D3DFVF_TEX1)
const int num_vertices = 4;
Vertex vertices[num_vertices] = 
{
    50.0f, 50.f, 0.5f, 1.0f, 0xFFFF0000, 0.0f, 0.0f, 
    450.0f, 50.f, 0.5f, 1.0f, 0xFFFF0000, 1.0f, 0.0f, 
    450.0f, 450.f, 0.5f, 1.0f, 0xFFFF0000, 1.0f, 1.0f,
    50.0f, 450.0f, 0.05f, 1.0f, 0xFFFF0000, 0.0f, 1.0f
};

/*+-------------------------------------------------------------------
  Global Variables

  These are global variables so that I can be guaranteed that 
  these interfaces will be cleaned up on exit.  
*/
IDirect3D9* g_lpD3D = NULL;
IDirect3DDevice9* g_lpDevice = NULL;
IDirect3DVertexBuffer9* g_lpVB = NULL;
IDirect3DTexture9* g_lpTexture = NULL;
IDirect3DPixelShader9* g_lpPixelShader = NULL;


/*+-------------------------------------------------------------------
  check_device_capabilities 

  Check to see if the specified device has the necessary
  capabilities for this application, i.e. that it supports
  pixel shader version 2.0.

*/
bool check_device_capabilities (IDirect3D9* lpD3D, D3DDEVTYPE dev_type)
{
    assert(lpD3D != NULL);

    D3DCAPS9 d3dcaps9;
    ZeroMemory(&d3dcaps9, sizeof(d3dcaps9));

    HRESULT hRes = lpD3D->GetDeviceCaps(D3DADAPTER_DEFAULT,
					dev_type,
					&d3dcaps9);
    

    // GetDeviceCaps with DX9 for D3DDEVTYPE_HAL can fail 
    // on older graphics cards such as the TNT2.  
    if (FAILED(hRes)) return false; 

    // Check for pixel shader version 2.0 support.
    if (d3dcaps9.PixelShaderVersion < D3DPS_VERSION(2,0))
        return false;


    return true;
}


/*+-------------------------------------------------------------------
  init_dx9 
  
  Initialize DirectX 9, set the IDirect3D9 and IDirect3DDevice9 
  global variables.  Also check for support for floating point
  textures, and for pixel shaders. 
  
*/
bool init_dx9 (HWND hwnd)
{
    assert(hwnd != NULL);
    g_lpD3D = Direct3DCreate9(D3D_SDK_VERSION);

    if (!g_lpD3D) return false; 

    D3DDEVTYPE dev_type = D3DDEVTYPE_HAL;
    // Check to see if the HAL has the right capabilities. 
    if (!check_device_capabilities(g_lpD3D, 
				   dev_type))
    {
	// If the HAL device doesn't support ps 2.0, then use the REF
	// device.  This device is much slower, but it's fully
	// featured.
	
        dev_type = D3DDEVTYPE_REF;
    }
    
    // Get the current display mode.  It's needed for
    // the call to IDirect3D9::CreateDevice. 
    D3DDISPLAYMODE displayMode;
    ZeroMemory(&displayMode, sizeof(displayMode));
    
    HRESULT hRes = g_lpD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT,
                                                  &displayMode);
    
    if (FAILED(hRes)) return false;

    D3DPRESENT_PARAMETERS d3dpp;
    ZeroMemory(&d3dpp, sizeof(d3dpp));

    d3dpp.BackBufferFormat = displayMode.Format;
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D24S8;

    hRes = g_lpD3D->CreateDevice(D3DADAPTER_DEFAULT,
                                 dev_type,
                                 hwnd,
                                 // FIXME - Is this flag right? 
                                 D3DCREATE_SOFTWARE_VERTEXPROCESSING, 
                                 &d3dpp,
                                 &g_lpDevice);

    if (FAILED(hRes)) return false;
                                 
    // Load a texture from a test.bmp file. 
    hRes = D3DXCreateTextureFromFile(g_lpDevice, 
                                     "test.jpg",
                                     &g_lpTexture);
 
    if (FAILED(hRes)) return false;
    
    // We're done with the texture interface, so release
    // it. 
    hRes = g_lpDevice->SetTexture(0, g_lpTexture);
    if (FAILED(hRes)) return false;

    // We're done with the texture, so release it. 
    g_lpTexture->Release();
    g_lpTexture = NULL;

    hRes = g_lpDevice->CreateVertexBuffer(num_vertices * sizeof(Vertex),
                                          D3DUSAGE_WRITEONLY, 
                                          VERTEX_FORMAT, 
                                          D3DPOOL_DEFAULT,
                                          &g_lpVB, NULL);
    
    if (FAILED(hRes)) return false;

    // Lock the vertex buffer and populate it with the 
    // vertex data. 
    VOID* ptr = NULL;
    
    hRes = g_lpVB->Lock(0, num_vertices * sizeof(Vertex), &ptr, 0);
    if (FAILED(hRes)) return false;

    memcpy(ptr, vertices, num_vertices * sizeof(Vertex));

    hRes = g_lpVB->Unlock();
    if (FAILED(hRes)) return false;
    
    hRes = g_lpDevice->SetFVF(VERTEX_FORMAT);
    if (FAILED(hRes)) return false;

    hRes = g_lpDevice->SetStreamSource(0, g_lpVB, 0, sizeof(Vertex));
    if (FAILED(hRes)) return false;

    // Assemble, create, and set the pixel shader 
    LPD3DXBUFFER pCode = NULL;
    LPD3DXBUFFER pErrorMsgs = NULL;

    hRes = D3DXAssembleShaderFromFile("simple_texture_map.ps", NULL, NULL, 0,
                                      &pCode, &pErrorMsgs);

    // If assembly failed, then pErrorMsgs will hopefully be non-null
    // and it will contain some useful error message from the
    // assembler.  Setting a breakpoint on the "return false" statement
    // is probably a good idea while you're debugging. 
    if ((FAILED(hRes)) && (pErrorMsgs != NULL))
    {
        unsigned char* message = 
	    (unsigned char*)pErrorMsgs->GetBufferPointer();

        return false;
    }

    // Create the pixel shader
    hRes = g_lpDevice->CreatePixelShader((DWORD*)pCode->GetBufferPointer(), 
                                         &g_lpPixelShader);
    if (FAILED(hRes)) return false;


    // Now set the pixel shader
    hRes = g_lpDevice->SetPixelShader(g_lpPixelShader);
    if (FAILED(hRes)) return false;
    
    // Done with the pixel shader interface, so release it.
    g_lpPixelShader->Release();
    g_lpPixelShader = NULL;

    // There's no lighting in this example. 
    g_lpDevice->SetRenderState(D3DRS_LIGHTING, FALSE);

    // Return true to indicate that initialization succeeded.
    return true;
}


/*+-------------------------------------------------------------------
  cleanup 

  This function runs just before the program exits.  This is
  where DX or other system resources should be cleaned up.

*/
void cleanup (void)
{
    if (g_lpTexture)
    {
	// g_lpTexture has already been released and NULLed, and so it
	// should be NULL, but just in case ...
	g_lpTexture->Release();
	g_lpTexture = NULL;
    }
    if (g_lpPixelShader)
    {
	// g_lpPixelShader has already been released and NULLed, so 
	// it should be NULL at this point, but just in case ...
	
	g_lpPixelShader->Release();
	g_lpPixelShader = NULL;
    }
    if (g_lpVB)
    {
        g_lpVB->Release();
        g_lpVB = NULL;
    }

    if (g_lpDevice)
    {
        g_lpDevice->Release();
        g_lpDevice = NULL;
    }

    if (g_lpD3D)
    {
        g_lpD3D->Release();
        g_lpD3D = NULL;
    }
}



/*+-------------------------------------------------------------------
  render_scene (function)

  Use the global DX9 device (g_lpDevice) to render the scene to the
  window.

*/
void render_scene (void)
{
    assert(g_lpDevice != NULL);

    HRESULT hRes = 
        g_lpDevice->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 
                          D3DCOLOR_COLORVALUE(1.0, 1.0, 1.0, 1.0), // white
                          1.0, // depth clear value
                          0);
    
    g_lpDevice->BeginScene();

    g_lpDevice->DrawPrimitive(D3DPT_TRIANGLEFAN, 0, 2);

    g_lpDevice->EndScene();

    hRes = g_lpDevice->Present(NULL, NULL, NULL, NULL);
}

/*+-------------------------------------------------------------------
      MainWndProc (callback function)

*/
LRESULT CALLBACK MainWndProc (HWND hwnd, UINT message, 
                              WPARAM wParam, LPARAM lParam)
{
    switch(message)
    {
    case WM_PAINT:
        render_scene();
        ValidateRect(hwnd, NULL);
        return 0L;

    case WM_DESTROY:
        PostQuitMessage(0);
        return 0L;
    }

    return DefWindowProc(hwnd, message, wParam, lParam);
}

int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE /* hPrevInstance */,
                    LPSTR lpszCmdLine, int nCmdShow)
{
    MSG msg;
    TCHAR szClassName[] = TEXT("ps2t");
    WNDCLASSEX wndclassx;

    wndclassx.cbSize = sizeof(WNDCLASSEX);
    wndclassx.style = CS_HREDRAW | CS_VREDRAW;
    wndclassx.lpfnWndProc = MainWndProc;
    wndclassx.cbClsExtra = 0;
    wndclassx.cbWndExtra = 0;
    wndclassx.hInstance = hInstance;
    wndclassx.hIcon = LoadIcon(NULL, IDI_APPLICATION);
    wndclassx.hCursor = LoadCursor(NULL, IDC_ARROW);
    wndclassx.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
    wndclassx.lpszMenuName = NULL;
    wndclassx.lpszClassName = szClassName;
    wndclassx.hIconSm = NULL;
    
    ATOM aRC = RegisterClassEx(&wndclassx);

    HWND hwnd = CreateWindowEx(WS_EX_APPWINDOW,
                               szClassName,
                               TEXT("A Simple Pixel Shader 2.0 Example"),
                               WS_OVERLAPPEDWINDOW,
                               CW_USEDEFAULT,
                               CW_USEDEFAULT,
                               500,  // width
                               500,  // height
                               NULL,
                               NULL,
                               hInstance,
                               NULL);

    if (init_dx9(hwnd))
    {
        ShowWindow(hwnd, nCmdShow);
        UpdateWindow(hwnd);
        
        while (GetMessage(&msg, 0, 0, 0))
        {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    }

    cleanup();

    return msg.wParam;
}
















